from baby_steps import given, then, when
from district42 import optional, schema

from schemax import from_json_schema


def test_empty_schema():
    with given:
        jsch = {}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.any


def test_none():
    with given:
        jsch = {"type": "null"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.none


def test_bool():
    with given:
        jsch = {"type": "boolean"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.bool


def test_bool_with_value():
    with given:
        jsch = {"enum": [True]}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.bool(True)


def test_int():
    with given:
        jsch = {"type": "integer"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int


def test_int_with_value():
    with given:
        jsch = {"type": "integer", "minimum": 3, "maximum": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int(3)


def test_int_with_min():
    with given:
        jsch = {"type": "integer", "minimum": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int.min(3)


def test_int_with_exclusive_min():
    with given:
        jsch = {"type": "integer", "exclusiveMinimum": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int.min(4)


def test_int_with_max():
    with given:
        jsch = {"type": "integer", "maximum": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int.max(3)


def test_int_with_exclusive_max():
    with given:
        jsch = {"type": "integer", "exclusiveMaximum": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int.max(2)


def test_int_with_min_max():
    with given:
        jsch = {"type": "integer", "minimum": 3, "maximum": 4}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int.min(3).max(4)


def test_int_with_exclusive_min_max():
    with given:
        jsch = {"type": "integer", "exclusiveMinimum": 1, "exclusiveMaximum": 4}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int.min(2).max(3)


def test_float():
    with given:
        jsch = {"type": "number"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.float


def test_float_with_value():
    with given:
        jsch = {"type": "number", "minimum": 3.14, "maximum": 3.14}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.float(3.14)


def test_float_with_min():
    with given:
        jsch = {"type": "number", "minimum": 3.14}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.float.min(3.14)


def test_float_with_max():
    with given:
        jsch = {"type": "number", "maximum": 3.14}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.float.max(3.14)


def test_float_with_min_max():
    with given:
        jsch = {"type": "number", "minimum": 3.14, "maximum": 4.15}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.float.min(3.14).max(4.15)


def test_str():
    with given:
        jsch = {"type": "string"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.str


def test_str_with_min():
    with given:
        jsch = {"type": "string", "minLength": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.str.len(3, ...)


def test_str_with_max():
    with given:
        jsch = {"type": "string", "maxLength": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.str.len(..., 3)


def test_str_with_min_max():
    with given:
        jsch = {"type": "string", "minLength": 3, "maxLength": 14}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.str.len(3, 14)


def test_str_with_same_min_max():
    with given:
        jsch = {"type": "string", "minLength": 3, "maxLength": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.str.len(3)


def test_str_with_pattern_to_pattern():
    with given:
        jsch = {"type": "string", "pattern": "a+(b|c)+"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.str.regex("a+(b|c)+")


def test_list():
    with given:
        jsch = {"type": "array"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list


def test_list_with_elements():
    with given:
        jsch = {
            "type": "array",
            "prefixItems": [{"type": "string"}, {"type": "integer"}],
            "items": False
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list([schema.str, schema.int])


def test_list_with_elements_and_additions():
    with given:
        jsch = {
            "type": "array",
            "prefixItems": [{"type": "string"}, {"type": "integer"}],
            "items": True
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list([schema.str, schema.int, ...])


def test_list_with_type_contains():
    with given:
        jsch = {"type": "array", "contains": {"type": "string"}}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.unordered([..., schema.str, ...])


def test_list_with_type_items():
    with given:
        jsch = {"type": "array", "items": {"type": "string"}}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list(schema.str)


def test_list_with_min():
    with given:
        jsch = {"type": "array", "minItems": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list.len(3)


def test_list_with_max():
    with given:
        jsch = {"type": "array", "maxItems": 3}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list.len(..., 3)


def test_list_with_min_max():
    with given:
        jsch = {"type": "array", "minItems": 3, "maxItems": 14}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list.len(3, 14)


def test_object():
    with given:
        jsch = {"type": "object"}
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict


def test_object_with_required_key():
    with given:
        jsch = {
            "type": "object",
            "properties": {"list": {"type": "array"}},
            "required": ["list"],
            "additionalProperties": False
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict({"list": schema.list})


def test_object_with_required_key_and_additions():
    with given:
        jsch = {
            "type": "object",
            "properties": {"list": {"type": "array"}},
            "required": ["list"],
            "additionalProperties": True
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict({"list": schema.list, ...: ...})


def test_object_with_required_key_and_additions_2():
    with given:
        jsch = {
            "type": "object",
            "properties": {"list": {"type": "array"}},
            "required": ["list"]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict({"list": schema.list, ...: ...})


def test_object_with_optional_key():
    with given:
        jsch = {
            "type": "object",
            "properties": {"list": {"type": "array"}},
            "additionalProperties": False
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict({optional("list"): schema.list})


def test_object_with_optional_key_and_additions():
    with given:
        jsch = {
            "type": "object",
            "properties": {"list": {"type": "array"}},
            "additionalProperties": True
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict({optional("list"): schema.list, ...: ...})


def test_object_with_optional_key_and_additions_2():
    with given:
        jsch = {
            "type": "object",
            "properties": {"list": {"type": "array"}}
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict({optional("list"): schema.list, ...: ...})


def test_enum_none_value():
    with given:
        jsch = {
            "enum": [None]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.none


def test_enum_bool_value():
    with given:
        jsch = {
            "enum": [True]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.bool(True)


def test_enum_int_value():
    with given:
        jsch = {
            "enum": [123]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.int(123)


def test_enum_float_value():
    with given:
        jsch = {
            "enum": [3.1415]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.float(3.1415)


def test_enum_str_value():
    with given:
        jsch = {
            "enum": ["test"]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.str("test")


def test_enum_list_value():
    with given:
        jsch = {
            "enum": [["test"]]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.list


def test_enum_dict_value():
    with given:
        jsch = {
            "enum": [{"bob": "yes"}]
        }
    with when:
        res = from_json_schema(jsch)
    with then:
        assert res == schema.dict



import re
import warnings
from typing import Any, Dict

from district42 import SchemaVisitor
from district42.types import (
    AnySchema,
    BoolSchema,
    BytesSchema,
    ConstSchema,
    DictSchema,
    FloatSchema,
    GenericTypeAliasSchema,
    IntSchema,
    ListSchema,
    NoneSchema,
    StrSchema,
    TypeAliasPropsType,
)
from district42.utils import is_ellipsis
from niltype import Nil

from schemax import supported_props


class Translator(SchemaVisitor[Any]):
    def visit_none(self, schema: NoneSchema, **kwargs: Any) -> Dict[str, Any]:
        return {"type": "null"}

    def visit_bool(self, schema: BoolSchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.BoolProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        if schema.props.value is Nil:
            return {"type": "boolean"}

        return {"enum": [schema.props.value]}

    def visit_int(self, schema: IntSchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.IntProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        int_object: Dict[str, Any] = {
            "type": "integer"
        }

        if schema.props.value is not Nil:
            int_object["minimum"] = schema.props.value
            int_object["maximum"] = schema.props.value

        if schema.props.min is not Nil:
            int_object["minimum"] = schema.props.min
        if schema.props.max is not Nil:
            int_object["maximum"] = schema.props.max

        return int_object

    def visit_float(self, schema: FloatSchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.FloatProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        number_object: Dict[str, Any] = {
            "type": "number"
        }

        if schema.props.value is not Nil:
            number_object["minimum"] = schema.props.value
            number_object["maximum"] = schema.props.value

        if schema.props.min is not Nil:
            number_object["minimum"] = schema.props.min
        if schema.props.max is not Nil:
            number_object["maximum"] = schema.props.max

        return number_object

    def visit_str(self, schema: StrSchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.StrProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        str_object: Dict[str, Any] = {
            "type": "string"
        }

        if schema.props.value is not Nil:
            str_object["const"] = schema.props.value

        if schema.props.pattern is not Nil:
            if re.search(r"\\\w", schema.props.pattern) is not None:
                warnings.warn("Be aware that escape-sequences are unsupported in json-schemas "
                              "regexes. Currently we can't do reformation and provide them "
                              "'as it is'.\nUse at our own risk!", Warning)
            str_object["pattern"] = schema.props.pattern

        if schema.props.len is not Nil:
            str_object["minLength"] = schema.props.len
            str_object["maxLength"] = schema.props.len

        if schema.props.min_len is not Nil:
            str_object["minLength"] = schema.props.min_len
        if schema.props.max_len is not Nil:
            str_object["maxLength"] = schema.props.max_len

        if schema.props.alphabet is not Nil:
            str_object["pattern"] = "^(" \
                                    + "|".join(re.escape(a) for a in schema.props.alphabet) \
                                    + ")+$"

        if schema.props.substr is not Nil:
            str_object["pattern"] = f"^.*({re.escape(schema.props.substr)}).*$"

        return str_object

    def visit_list(self, schema: ListSchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.ListProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        array_object: Dict[str, Any] = {
            "type": "array"
        }

        if schema.props.len is not Nil:
            array_object["minItems"] = schema.props.len
            array_object["maxItems"] = schema.props.len

        if schema.props.min_len is not Nil:
            array_object["minItems"] = schema.props.min_len
        if schema.props.max_len is not Nil:
            array_object["maxItems"] = schema.props.max_len

        if schema.props.type is not Nil:
            array_object["items"] = schema.props.type.__accept__(self, **kwargs)
            return array_object

        if schema.props.elements is not Nil:
            array_object["prefixItems"] = []
            array_object["items"] = False
            for element in schema.props.elements:
                if is_ellipsis(element):
                    array_object["items"] = True
                    continue

                array_object["prefixItems"].append(element.__accept__(self, **kwargs))

        return array_object

    def visit_dict(self, schema: DictSchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.DictProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        dict_object: Dict[str, Any] = {
            "type": "object"
        }

        if schema.props.keys is Nil:
            return dict_object

        dict_object["additionalProperties"] = False
        dict_object["properties"] = {}
        required = []
        for key, (val, is_optional) in schema.props.keys.items():
            if is_ellipsis(key):
                dict_object["additionalProperties"] = True
                continue

            dict_object["properties"][key] = val.__accept__(self, **kwargs)
            if not is_optional:
                required.append(key)

        if required:
            dict_object["required"] = required

        return dict_object

    def visit_any(self, schema: AnySchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.AnyProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        any_of = []

        if schema.props.types is not Nil:
            for obj in schema.props.types:
                any_of.append(obj.__accept__(self))

        return {"anyOf": any_of}

    def visit_const(self, schema: ConstSchema, **kwargs: Any) -> Dict[str, Any]:
        for prop in schema.props:
            if prop not in supported_props.ConstProps:
                warnings.warn(f"Unsupported prop {prop} for type {schema.__str__()}", Warning)

        return {"const": schema.props.value}

    def visit_bytes(self, schema: BytesSchema, **kwargs: Any) -> Dict[str, Any]:
        warnings.warn("'schema.bytes' is not implemented")
        return {}

    def visit_type_alias(self, schema: GenericTypeAliasSchema[TypeAliasPropsType],
                         **kwargs: Any) -> Any:
        warnings.warn("'schema.alias' is not implemented")
        return {}



import argparse
import json
from json import JSONDecodeError
from typing import Optional

import yaml

from schemax import from_json_schema
from schemax._data_collector import collect_schema_data
from schemax._generator import MainGenerator


def translate(files: str) -> None:
    if len(files) < 1:
        print("Error: No JSON-Schema file provided.")
        exit(1)

    for file in files:
        print(f"Translation from JSON-Schema to d42-schema for '{file}':")
        try:
            with open(file, "r") as f:
                print(from_json_schema(json.load(f)), end="\n\n")
        except FileNotFoundError:
            print("File doesn't exist", end="\n\n")
            continue
        except JSONDecodeError:
            print("File doesn't contain JSON", end='\n\n')
            continue


def generate(file: str, base_url: Optional[str] = None, humanize: bool = False) -> None:
    if len(file) < 1:
        print("No OpenAPI file provided")
        exit(1)
    try:
        with open(file, "r") as f:
            print("Generating schemas and interfaces from given OpenApi...")
            if f.name.endswith(".json"):
                schema_data = collect_schema_data(json.load(f))
            elif f.name.endswith(".yaml") or f.name.endswith(".yml"):
                schema_data = collect_schema_data(yaml.load(f, yaml.FullLoader))
            else:
                print(f"'{f.name}' type is not .json or .yaml file")
                exit(1)

            generator = MainGenerator(schema_data)
            generator.all()
            print("Successfully generated")
    except FileNotFoundError:
        print(f"File '{file}' doesn't exist")
        exit(1)
    except JSONDecodeError:
        print(f"File '{f.name}' doesn't contain JSON")
        exit(1)


def main() -> None:
    parser = argparse.ArgumentParser(
        description="CLI script with generate and translate functions")

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Команда generate
    generate_parser = subparsers.add_parser("generate", help="Generate from a file")
    generate_parser.add_argument("input_file", help="Input file for generation")
    generate_parser.add_argument("--base_url", help="Base API URL for interface")
    generate_parser.add_argument(
        "--humanize", action="store_true", help="Make interface human-readable")

    # Команда translate
    translate_parser = subparsers.add_parser("translate", help="Translate from multiple files")
    translate_parser.add_argument("input_files", nargs="+", help="Input files for translation")

    args = parser.parse_args()

    if args.command == "generate":
        generate(args.input_file, args.base_url, args.humanize)
    elif args.command == "translate":
        translate(args.input_files)
    else:
        print("Unknown command")
        parser.print_help()


if __name__ == '__main__':
    main()



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import jsonschema
from district42 import optional
from district42.types import (
    AnySchema,
    BoolSchema,
    DictSchema,
    FloatSchema,
    GenericSchema,
    IntSchema,
    ListSchema,
    NoneSchema,
    StrSchema,
)
from district42_exp_types.unordered import UnorderedSchema

if TYPE_CHECKING:
    import builtins

    EllipsisType = builtins.ellipsis
else:
    EllipsisType = Any


def null_visitor() -> NoneSchema:
    return NoneSchema()


def boolean_visitor(value: Optional[bool] = None) -> BoolSchema:
    if value is None:
        return BoolSchema()
    if value is True:
        return BoolSchema()(True)
    if value is False:
        return BoolSchema()(False)

    return BoolSchema()


def integer_visitor(value: Dict[str, Any]) -> IntSchema:
    sch = IntSchema()

    if "maximum" in value and "minimum" in value:
        if value["minimum"] != value["maximum"]:
            return sch.min(value["minimum"]).max(value["maximum"])
        return sch(value["minimum"])

    if "minimum" in value:
        sch = sch.min(value["minimum"])

    if "maximum" in value:
        sch = sch.max(value["maximum"])

    if "exclusiveMinimum" in value:
        sch = sch.min(value["exclusiveMinimum"] + 1)

    if "exclusiveMaximum" in value:
        sch = sch.max(value["exclusiveMaximum"] - 1)

    return sch


def number_visitor(value: Dict[str, Any]) -> FloatSchema:
    sch = FloatSchema()

    if "maximum" in value and "minimum" in value:
        if value["minimum"] != value["maximum"]:
            return sch.min(float(value["minimum"])).max(float(value["maximum"]))
        return sch(float(value["minimum"]))

    if "minimum" in value:
        sch = sch.min(float(value["minimum"]))

    if "maximum" in value:
        sch = sch.max(float(value["maximum"]))

    return sch


def string_visitor(value: Dict[str, Any]) -> StrSchema:
    sch = StrSchema()

    if "minLength" in value and "maxLength" in value:
        if value["minLength"] == value["maxLength"]:
            return sch.len(value["minLength"])
        return sch.len(value["minLength"], value["maxLength"])

    if "minLength" in value:
        return sch.len(value["minLength"], ...)

    if "maxLength" in value:
        return sch.len(..., value["maxLength"])

    if "pattern" in value:
        sch = sch.regex(value["pattern"])

    return sch


def array_visitor(value: Dict[str, Any]) -> ListSchema:
    sch = ListSchema()

    if "contains" in value:
        prop = _from_json_schema(value["contains"])
        sch = UnorderedSchema()([..., prop, ...])

    if "items" in value:
        if not isinstance(value["items"], bool):
            if "oneOf" in value["items"]:
                props = [_from_json_schema(item) for item in value["items"]["oneOf"]]
                sch = sch(props)  # type: ignore
            else:
                prop = _from_json_schema(value["items"])
                sch = sch(prop)

    if "prefixItems" in value:
        props = [_from_json_schema(item) for item in value["prefixItems"]]

        if value.get("items", True) is True:
            props.append(Ellipsis)  # type: ignore
        sch = sch(props)  # type: ignore

    if "minItems" in value and "maxItems" in value:
        if value["minItems"] != value["maxItems"]:
            return sch.len(value["minItems"], value["maxItems"])
        return sch.len(value["minItems"])

    if "minItems" in value:
        sch = sch.len(value["minItems"])

    if "maxItems" in value:
        sch = sch.len(..., value["maxItems"])

    return sch


def object_visitor(value: Dict[str, Any]) -> DictSchema:
    if "properties" not in value:
        return DictSchema()

    props: Dict[Any, Union[GenericSchema, EllipsisType]] = {}
    for key in value["properties"]:
        if "required" in value:
            if key in value["required"]:
                props[key] = _from_json_schema(value["properties"][key])
            else:
                props[optional(key)] = _from_json_schema(value["properties"][key])
        else:
            props[optional(key)] = _from_json_schema(value["properties"][key])

    if value.get("additionalProperties", True) is True:
        props[Ellipsis] = Ellipsis

    return DictSchema()(props)


def schema_normalize(value: Dict[str, Any]) -> Dict[str, Any]:
    def schema_runner(
        schema: Dict[str, Any], resolver: jsonschema.RefResolver
    ) -> Dict[str, Any]:
        if isinstance(schema, dict):
            if "$ref" in schema:
                with resolver.resolving(schema["$ref"]) as resolved:
                    return schema_runner(resolved, resolver)
            else:
                return {k: schema_runner(v, resolver) for k, v in schema.items()}
        elif isinstance(schema, list):
            return [schema_runner(item, resolver) for item in schema]  # noqa
        else:
            return schema

    ref_resolver = jsonschema.RefResolver("", value)
    return schema_runner(value, ref_resolver)


def _from_json_schema(value: Dict[Any, Any]) -> GenericSchema:
    # Dirty-dirty hack for OApi schemas, need somehow to clarify better
    if "allOf" in value:
        return _from_json_schema(value["allOf"][-1])

    if "enum" in value:
        props: List[GenericSchema] = []
        for var in value["enum"]:
            match type(var).__name__:
                case "NoneType":
                    props.append(NoneSchema())
                case "bool":
                    props.append(BoolSchema()(var))
                case "int":
                    props.append(IntSchema()(var))
                case "float":
                    props.append(FloatSchema()(var))
                case "str":
                    props.append(StrSchema()(var))
                case "list":
                    props.append(ListSchema())
                case "dict":
                    props.append(DictSchema())
        # If we have only one prop type in result we don't need it in AnySchema
        return AnySchema()(*props) if len(props) > 1 else props[0]

    if "type" not in value:
        return AnySchema()

    if isinstance(value["type"], list):
        schemas = []
        for i in value["type"]:
            schemas.append(_from_json_schema({"type": i}))
        if IntSchema() in schemas:
            schemas.append(FloatSchema())
        if FloatSchema() in schemas:
            schemas.append(IntSchema())
        return AnySchema()(*schemas)

    match value["type"]:
        case "null":
            return null_visitor()
        case "boolean":
            return boolean_visitor()
        case "integer":
            return integer_visitor(value)
        case "number":
            return number_visitor(value)
        case "string":
            return string_visitor(value)
        case "array":
            return array_visitor(value)
        case "object":
            return object_visitor(value)
        case _:
            return AnySchema()



from typing import List

BoolProps: List[str] = ["value"]
IntProps: List[str] = ["value", "min", "max"]
FloatProps: List[str] = ["value", "min", "max"]
StrProps: List[str] = ["value", "pattern", "len", "min_len", "max_len", "alphabet", "substr"]
ListProps: List[str] = ["len", "min_len", "max_len", "type", "elements"]
DictProps: List[str] = ["keys"]
AnyProps: List[str] = ["types"]
ConstProps: List[str] = ["value"]



from typing import Any, Dict, Optional

from district42 import GenericSchema
from district42.types import Schema

from ._from_json_schema import _from_json_schema, schema_normalize
from ._translator import Translator
from ._version import version

__version__ = version
__all__ = ("Translator", "to_json_schema", "from_json_schema", "schema_normalize")

_translator = Translator()


def to_json_schema(
    schema: GenericSchema,
    title: Optional[str] = None,
    hide_draft: Optional[bool] = False,
    **kwargs: Any
) -> Any:
    translation = schema.__accept__(_translator, **kwargs)

    if title is not None:
        translation = {'title': title, **translation}

    if not hide_draft:
        # Use latest version of json-schema draft
        translation = {'$schema': "https://json-schema.org/draft/2020-12/schema#", **translation}

    return translation


def from_json_schema(value: Dict[Any, Any]) -> GenericSchema:
    normalized_value = schema_normalize(value)
    return _from_json_schema(normalized_value)


Schema.__override__(Schema.__invert__.__name__, to_json_schema)



import re
from typing import Any, Dict, List, Tuple, Union

from schemax import from_json_schema

from ._from_json_schema import schema_normalize


def collect_schema_data(value: Dict[str, Any]) -> List[Dict[str, Any]]:
    schema_data: List[Dict[str, Any]] = []
    normalized_schema: Dict[str, Any] = schema_normalize(value)

    if "paths" in normalized_schema:
        for path, path_data in normalized_schema["paths"].items():
            for http_method, method_data in path_data.items():
                request_schema, response_schema = get_request_response_schemas(method_data)

                args = get_path_arguments(path)
                if request_schema:
                    args.append("body")

                tags = method_data.get("tags", [])
                schema_data.append({
                    "http_method": http_method,
                    "path": convert_to_snake_case(path),
                    "args": args,
                    "queries": None,
                    "interface_method": get_interface_method_name(http_method, path),
                    "status": get_success_status(method_data),
                    "schema_prefix": get_schema_prefix(http_method, path),
                    "response_schema": from_json_schema(response_schema),
                    "request_schema": from_json_schema(request_schema),
                    "tags": tags
                })

    return schema_data


def get_request_response_schemas(
    method_data: Dict[str, Any]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    request_schema: Dict[str, Any] = {}
    response_schema: Dict[str, Any] = {}

    if "requestBody" in method_data:
        request_schema = get_request_schema(method_data["requestBody"])
    elif "parameters" in method_data:
        request_schema = get_request_schema_from_parameters(method_data["parameters"])

    if "responses" in method_data:
        response_schema = get_response_schema(method_data["responses"])

    return request_schema, response_schema


def get_request_schema(request_body: Dict[str, Any]) -> Dict[str, Any]:
    content = request_body.get("content", {})
    json_content = content.get("application/json", {}).get("schema", {})
    return json_content  # type: ignore


def get_request_schema_from_parameters(parameters: List[Dict[str, Any]]) -> Dict[str, Any]:
    for param in parameters:
        if param["in"] == "body":
            return param["schema"]  # type: ignore
    return {}


def get_response_schema(responses: Dict[str, Any]) -> Dict[str, Any]:
    for status, status_data in responses.items():
        if status == "200" or status == 200:  # type: ignore
            content = status_data.get("content", {})
            content_schema: Dict[str, Any] = content.get(next(iter(content)), {}).get("schema", {})
            return content_schema
    return {}


def get_path_arguments(path: str) -> List[str]:
    return [convert_to_snake_case(arg) for arg in re.findall(r"{([^}]+)}", path)]


def get_interface_method_name(http_method: str, path: str) -> str:
    return (
        http_method.lower() +
        "_".join(
            convert_to_snake_case(word)
            .replace("{", "")
            .replace("}", "")
            .replace("-", "_")
            .lower()
            for word in path.split("/")
        )
    )


def get_success_status(method_data: Dict[str, Any]) -> Union[str, int]:
    success_statuses = ["200", 200]
    for status in success_statuses:
        if status in method_data.get("responses", {}):
            return status  # type: ignore
    return ""


def get_schema_prefix(http_method: str, path: str) -> str:
    return (
        http_method.capitalize() +
        "".join(
            word
            .replace("{", "")
            .replace("}", "")
            .replace("-", "")
            .capitalize()
            for word in path.split("/")
        )
    )


def convert_to_snake_case(input_string: str) -> str:
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', input_string).lower()



from baby_steps import given, then, when
from district42 import optional, schema

from schemax import to_json_schema


def test_none():
    with given:
        sch = schema.none
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "null"}


def test_bool():
    with given:
        sch = schema.bool
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "boolean"}


def test_bool_with_value():
    with given:
        sch = schema.bool(True)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"enum": [True]}


def test_int():
    with given:
        sch = schema.int
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "integer"}


def test_int_with_value():
    with given:
        sch = schema.int(3)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "integer", "minimum": 3, "maximum": 3}


def test_int_with_min():
    with given:
        sch = schema.int.min(3)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "integer", "minimum": 3}


def test_int_with_max():
    with given:
        sch = schema.int.max(3)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "integer", "maximum": 3}


def test_int_with_min_max():
    with given:
        sch = schema.int.min(3).max(6)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "integer", "minimum": 3, "maximum": 6}


def test_float():
    with given:
        sch = schema.float
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "number"}


def test_float_with_value():
    with given:
        sch = schema.float(3.14)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "number", "minimum": 3.14, "maximum": 3.14}


def test_float_with_min():
    with given:
        sch = schema.float.min(3.14)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "number", "minimum": 3.14}


def test_float_with_max():
    with given:
        sch = schema.float.max(3.14)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "number", "maximum": 3.14}


def test_float_with_min_max():
    with given:
        sch = schema.float.min(3.14).max(6.28)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "number", "minimum": 3.14, "maximum": 6.28}


def test_str():
    with given:
        sch = schema.str
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string"}


def test_str_with_value():
    with given:
        sch = schema.str("test")
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "const": "test"}


def test_str_with_len():
    with given:
        sch = schema.str.len(3)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "minLength": 3, "maxLength": 3}


def test_str_with_min():
    with given:
        sch = schema.str.len(3, ...)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "minLength": 3}


def test_str_with_max():
    with given:
        sch = schema.str.len(..., 3)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "maxLength": 3}


def test_str_with_min_max():
    with given:
        sch = schema.str.len(3, 14)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "minLength": 3, "maxLength": 14}


def test_str_with_regex():
    with given:
        sch = schema.str.regex(r"[A-Z-_]+")
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "pattern": "[A-Z-_]+"}


def test_str_with_alphabet():
    with given:
        sch = schema.str.alphabet("abcd")
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "pattern": "^(a|b|c|d)+$"}


def test_str_with_contains():
    with given:
        sch = schema.str.contains("TEST")
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "string", "pattern": "^.*(TEST).*$"}


def test_list():
    with given:
        sch = schema.list
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "array"}


def test_list_with_type_value():
    with given:
        sch = schema.list(schema.int)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "array", "items": {"type": "integer"}}


def test_list_with_elements_value():
    with given:
        sch = schema.list([schema.int, schema.str])
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "type": "array",
            "prefixItems": [{"type": "integer"}, {"type": "string"}],
            "items": False
        }


def test_list_with_elements_value_and_ellipsis():
    with given:
        sch = schema.list([schema.int, ...])
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "type": "array",
            "prefixItems": [{"type": "integer"}],
            "items": True
        }


def test_list_with_min():
    with given:
        sch = schema.list.len(3, ...)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "array", "minItems": 3}


def test_list_with_max():
    with given:
        sch = schema.list.len(..., 3)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "array", "maxItems": 3}


def test_list_with_min_max():
    with given:
        sch = schema.list.len(3, 14)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "array", "minItems": 3, "maxItems": 14}


def test_dict():
    with given:
        sch = schema.dict
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"type": "object"}


def test_dict_with_value():
    with given:
        sch = schema.dict({"foo": schema.int})
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "type": "object",
            "properties": {"foo": {"type": "integer"}},
            "required": ["foo"],
            "additionalProperties": False
        }


def test_dict_with_value_and_ellipsis():
    with given:
        sch = schema.dict({"foo": schema.int, ...: ...})
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "type": "object",
            "properties": {"foo": {"type": "integer"}},
            "required": ["foo"],
            "additionalProperties": True
        }


def test_dict_with_optional_value():
    with given:
        sch = schema.dict({optional("foo"): schema.int})
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "type": "object",
            "properties": {"foo": {"type": "integer"}},
            "additionalProperties": False
        }


def test_dict_with_optional_value_and_ellipsis():
    with given:
        sch = schema.dict({optional("foo"): schema.int, ...: ...})
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "type": "object",
            "properties": {"foo": {"type": "integer"}},
            "additionalProperties": True
        }


def test_dict_with_value_and_properties():
    with given:
        sch = schema.dict({"foo": schema.int.min(3).max(14)})
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "type": "object",
            "properties": {"foo": {"type": "integer", "minimum": 3, "maximum": 14}},
            "required": ["foo"],
            "additionalProperties": False
        }


def test_any():
    with given:
        sch = schema.any
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"anyOf": []}


def test_any_with_values():
    with given:
        sch = schema.any(schema.int, schema.str)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {"anyOf": [{"type": "integer"}, {"type": "string"}]}


def test_or_operator_with_values():
    with given:
        sch = schema.str("test") | schema.int.min(3)
    with when:
        res = to_json_schema(sch, hide_draft=True)
    with then:
        assert res == {
            "anyOf": [
                {"type": "string", "const": "test"},
                {"type": "integer", "minimum": 3}
            ]
        }



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List

from jinja2 import Environment, FileSystemLoader, Template


class Generator(ABC):
    @abstractmethod
    def _get_template(self, template_name: str) -> Template:
        pass

    def _create_dir(self, dir_name: str) -> None:
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)

    def _create_package(self, dir_name: str) -> None:
        self._create_dir(dir_name=dir_name)
        init_file_path = f'{dir_name}/__init__.py'
        if not os.path.exists(init_file_path):
            Path(init_file_path).touch()

    def _generate_by_template(self, file_path: str, template_name: str, **kwargs: Any) -> None:
        if not os.path.exists(file_path):
            template = self._get_template(template_name=template_name)
            with open(file_path, 'w') as file:
                file.write(template.render(**kwargs))

    def append_string(self, lst: List[str], suffix: str) -> List[str]:
        return [f"{suffix}.{item}" for item in lst]


class MainGenerator(Generator):
    __PATH_TEMPLATES = os.path.dirname(os.path.realpath(__file__)) + '/templates'
    __TEMPLATE_SCHEMAS = 'schemas.py.j2'
    __TEMPLATE_SCHEMA_DEFINITION = 'schema_definition.py.j2'
    __TEMPLATE_INTERFACES = 'interfaces.py.j2'
    __TEMPLATE_API_ROUTE = 'api_route.py.j2'
    __TEMPLATE_SCENARIO = 'scenario.py.j2'

    __DIRECTORY_SCHEMAS = 'schemas'
    __DIRECTORY_INTERFACES = 'interfaces'
    __DIRECTORY_SCENARIOS = 'scenarios'

    __FILE_API_INTERFACE = 'api.py'
    __FILE_RESPONSE_SCHEMAS = 'response_schemas.py'
    __FILE_REQUEST_SCHEMAS = 'request_schemas.py'

    def __init__(self, schema_data: List[Dict[str, Any]]):
        super().__init__()
        self.schema_data = schema_data
        self.__templates = Environment(loader=FileSystemLoader(self.__PATH_TEMPLATES))
        self.__templates.filters["append_str"] = self.append_string

    def response_schemas(self) -> None:
        self._create_package(self.__DIRECTORY_SCHEMAS)
        self._generate_by_template(
            file_path=f'{self.__DIRECTORY_SCHEMAS}/{self.__FILE_RESPONSE_SCHEMAS}',
            template_name=self.__TEMPLATE_SCHEMAS)

        with open(f'{self.__DIRECTORY_SCHEMAS}/{self.__FILE_RESPONSE_SCHEMAS}', 'a') as file:
            for data_item in self.schema_data:
                template = self._get_template(self.__TEMPLATE_SCHEMA_DEFINITION)
                file.write(
                    template.render(
                        schema_name=f'{data_item["schema_prefix"]}' + "ResponseSchema",
                        schema_definition=data_item["response_schema"]
                    )
                )

    def request_schemas(self) -> None:
        self._create_package(self.__DIRECTORY_SCHEMAS)
        self._generate_by_template(
            file_path=f'{self.__DIRECTORY_SCHEMAS}/{self.__FILE_REQUEST_SCHEMAS}',
            template_name=self.__TEMPLATE_SCHEMAS)

        with open(f'{self.__DIRECTORY_SCHEMAS}/{self.__FILE_REQUEST_SCHEMAS}', 'a') as file:
            for data_item in self.schema_data:
                if data_item["request_schema"] != "schema.any":
                    template = self._get_template(self.__TEMPLATE_SCHEMA_DEFINITION)
                    file.write(
                        template.render(
                            schema_name=f'{data_item["schema_prefix"]}' + "RequestSchema",
                            schema_definition=data_item["request_schema"]
                        )
                    )

    def interfaces(self) -> None:
        self._create_package(self.__DIRECTORY_INTERFACES)
        self._generate_by_template(
            file_path=f"{self.__DIRECTORY_INTERFACES}/{self.__FILE_API_INTERFACE}",
            template_name=self.__TEMPLATE_INTERFACES
        )

        with open(f"{self.__DIRECTORY_INTERFACES}/{self.__FILE_API_INTERFACE}", "a") as file:
            for data_item in self.schema_data:
                template = self._get_template(self.__TEMPLATE_API_ROUTE)
                file.write(
                    template.render(
                        interface_method=data_item["interface_method"],
                        http_method=data_item["http_method"].upper(),
                        path=data_item["path"],
                        args=data_item["args"],
                        request_schema=(
                            data_item["request_schema"]
                            if data_item["request_schema"] != "schema.any"
                            else None
                        )
                    )
                )

    def scenarios(self) -> None:
        self._create_package(self.__DIRECTORY_SCENARIOS)
        for data_item in self.schema_data:
            self._generate_by_template(
                file_path=f"{self.__DIRECTORY_SCENARIOS}/{data_item['interface_method']}.py",
                template_name=self.__TEMPLATE_SCENARIO,
                subject=data_item["interface_method"].split("_"),
                interface_method=data_item["interface_method"],
                args=data_item["args"],
                response_schema=data_item["schema_prefix"]+"ResponseSchema",
                request_schema=(
                    data_item["schema_prefix"]+"RequestSchema"
                    if data_item["request_schema"] != "schema.any"
                    else None
                )
            )

    def all(self) -> None:
        self.request_schemas()
        self.response_schemas()
        self.interfaces()
        self.scenarios()

    def _get_template(self, template_name: str) -> Template:
        return self.__templates.get_template(name=template_name)
